/*
Copyright (C) 2006 by Intra2net AG - Gerd v. Egidy

The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/epoll.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include <sstream>
#include <iostream>
#include <algorithm>

#include "socket_handler.hxx"
#include "t2n_exception.hxx"
#include "log.hxx"

using namespace std;

namespace libt2n
{

socket_handler::socket_handler(int _sock, socket_type_value _socket_type)
    : sock(_sock)
    , recv_buffer_size( default_recv_buffer_size )
    , write_block_size( default_write_block_size )
    , write_timeout( default_write_timeout )
    , socket_type(_socket_type)
{
}

/**
 * Destructor. Closes open socket
 */
socket_handler::~socket_handler()
{
    if (sock != -1)
    {
        shutdown(sock,SHUT_RDWR);
        remove_from_epoll(sock);
        ::close(sock);

        sock = -1;
    }
}

/// close the underlying socket connection. Don't call directly, use the version provided
/// by the connection class you are using.
void socket_handler::close()
{
    if (sock != -1)
    {
        LOGSTREAM(debug, "close connection");
        // graceful shutdown
        shutdown(sock, SHUT_RDWR);\
        //taking it out of the epoll pool

        remove_from_epoll(sock);
        ::close(sock);

        sock = -1;
    }
}

/// set options like fast reuse and keepalive every socket should have
void socket_handler::set_socket_options(int sock)
{
    int i=1;

    /* fast reuse enable */
    if (setsockopt(sock,SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"error setting socket option: " << strerror(errno));

    /* keepalive enable */
    if (setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE, &i, sizeof(i)) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"error setting socket option: " << strerror(errno));

    /* close on exec */
    int fdflags;
    fdflags=fcntl(sock,F_GETFD, 0);
    if (fdflags < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    fdflags |= FD_CLOEXEC;
    if (fcntl(sock,F_SETFD,fdflags) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    /* non-blocking mode */
    int flflags;
    flflags=fcntl(sock,F_GETFL,0);
    if (flflags < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    flflags |= O_NONBLOCK;
    if (fcntl(sock,F_SETFL,flflags) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));
}

/// is the underlying socket connection still open?
bool socket_handler::is_closed()
{
    int r=fcntl(sock,F_GETFL);

    return !(r & O_ACCMODE);
}


/**
 * @brief set a new size for the receive buffer.
 * @param new_recv_buffer_size the new size for the receive buffer.
 *
 * The receive buffer determines the amount of data which is tried to read at once
 * from the underlying socket.
 *
 * The value is normalized to be at least 512 bytes and at max 32K bytes.
 */
void socket_handler::set_recv_buffer_size(unsigned int new_recv_buffer_size)
{
    recv_buffer_size= std::max( 512u, std::min( 32u * 1024u, new_recv_buffer_size ));
} //


/**
 * @brief set new size for the data chunks when writing.
 * @param new_write_block_size the new chunk size.
 *
 * The write block size determines the amount of data which is tried to write
 * to the socket when data needs to be sent.
 * Since writing data is done in a loop, this does not limit the amount of data which can
 * be written.
 *
 * The value is normalized to be at least 512 bytes and at max 32K bytes.
 */
void socket_handler::set_write_block_size(unsigned int new_write_block_size)
{
    write_block_size= std::max( 512u, std::min( 32u * 1024u, new_write_block_size ));
} //


/**
 * @brief set new timeout for writing a block
 * @param new_write_timeout the new timeout in usecs, -1: wait endless
 *
 * The write timeout determines the maximum amount of time that is waited
 * between writing each block. If the timeout is exceeded, write will
 * throw t2n_transfer_error
 */
void socket_handler::set_write_timeout(long long new_write_timeout)
{
    write_block_size=new_write_timeout;
} //


/** @brief check if new data is waiting on the raw socket
    @param[in,out] timeout wait until new data is found, max timeout millisecs.
            -1: wait endless
            0: return instantly
*/
bool socket_handler::data_waiting(int timeout)
{
    int ret = epoll_wait(epoll_fd,epoll_events,EPOLL_MAX_EVENTS,timeout);

    if (ret > 0)
        return true;
    else
        return false;
}

/** @brief read data from the raw socket and copy it into the provided buffer
    @param buffer the buffer where to append the new data
    @param[in,out] timeout wait until new data is found, max timeout millisecs.
            -1: wait endless
            0: return instantly
*/
bool socket_handler::fill_buffer(std::string& buffer,int timeout)
{
    // fast path for timeout==0
    if (timeout == 0 || data_waiting(timeout))
        return fill_buffer(buffer);
    else
        return false;
}

/** @brief read data from the raw socket and copy it into the provided buffer. Returns
           instantly if no data is waiting.
    @param buffer the buffer where to append the new data
*/
bool socket_handler::fill_buffer(std::string& buffer)
{
    char socket_buffer[recv_buffer_size];

    const int loop_max = 32;            /* limit is 32 * (default) 2048 bytes -> 65536 bytes in one go */
    int loop_count = 0;

    bool read_something = false;
    while (loop_count < loop_max)
    {
        const int nbytes = read(sock, socket_buffer, recv_buffer_size);

        if (nbytes < 0)
        {
            if (errno == EAGAIN || errno == EWOULDBLOCK)
                return read_something;                // no (more) data was waiting
            else if (errno == EINTR)
            {
                // interrupted, try again
                LOGSTREAM(debug, "EINTR received on read(), trying again");
            } else
                EXCEPTIONSTREAM(error, t2n_transfer_error, "error reading from socket : " << strerror(errno));
        }

        // End-of-file
        if (nbytes == 0)
        {
            LOGSTREAM(debug, "0 bytes received on read(), closing connection");
            close();
            return read_something;
        }

        // Data read -> store it
        if (nbytes > 0)
        {
            buffer.append(socket_buffer, nbytes);
            LOGSTREAM(debug, nbytes << " bytes read");
            read_something = true;
        }

        // more data waiting -> loop once more (up to loop_max)
        if (data_waiting(0))
        {
            ++loop_count;
        } else
        {
            break;
        }
    }

    return read_something;
}

/// writes raw data to the socket. Don't use directly, use the write() function provided by the 
/// connection because it encapsulates the data.
void socket_handler::socket_write(const std::string& data)
{
    unsigned int offset = 0;
    while (offset < data.size())
    {
        unsigned int write_size=write_block_size;

        if (offset+write_size > data.size())
            write_size = data.size()-offset;

        int rtn;
        while ((rtn=::write(sock, data.data()+offset, write_size)) == -1 &&
               (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR))
        {
            wait_ready_to_write(write_timeout);
            LOGSTREAM(debug,"resuming write() call after EAGAIN or EINTR or EWOULDBLOCK");
        }

        if (rtn == -1)
            EXCEPTIONSTREAM(error,t2n_transfer_error,"write() returned " << strerror(errno));
        else if (rtn != write_size)
        {
            LOGSTREAM(debug,"write() wrote " << rtn << " bytes, should have been "
                << write_size << " (complete: " << data.size() << ")");

            write_size = rtn;
        }

        offset += write_size;
    }

    LOGSTREAM(debug,"wrote " << data.size() << " bytes");
} // eo socket_handler::socket_write(const std::string&)

/// wait until the socket is ready to write again
void socket_handler::wait_ready_to_write(int write_block_timeout)
{
    // let's wait for the socket to become writable again...
    int rtn;
    while ((rtn=::epoll_wait(epoll_fd,epoll_events,EPOLL_MAX_EVENTS,write_block_timeout)) ==-1 && errno == EINTR);

    if (rtn > 0)
    {
        for (int i=0; i < rtn; ++i)
        {
            // Something very wrong happened
            if(epoll_events[i].events != EPOLLOUT && epoll_events[i].events == EPOLLERR)
                EXCEPTIONSTREAM(error,t2n_transfer_error,"exception on socket; cannot write any more.");
        }
    }
    else if (rtn == 0)
        EXCEPTIONSTREAM(error,t2n_transfer_error,"timeout on epoll_wait() for write");
    else if (rtn == -1)
        EXCEPTIONSTREAM(error,t2n_transfer_error,"cannot find socket to write: " << strerror(errno));
}

/**
 * @brief  initialize the epoll pool and add the primary socket to it
 * assuming the socket is already created!
 */
void socket_handler::create_epoll()
{
    struct epoll_event ev = {};
    ev.events = EPOLLIN;
    epoll_fd = epoll_create1(EPOLL_CLOEXEC);

    if (epoll_fd == -1)
        EXCEPTIONSTREAM(error,t2n_server_error, "error opening epoll: " << strerror(errno));

    if(sock != -1)
        add_to_epoll(sock,&sock);
    else
        EXCEPTIONSTREAM(error,t2n_server_error, "socket not previously created impossible to add to epoll");
}

/**
 * @brief this adds one file descriptor to the epoll pool to monitor events
 * assuming the epoll was previously created
 * @param fd the file descriptor we want to monitor
 */
void socket_handler::add_to_epoll(int fd,void* ptr)
{
    struct epoll_event ev;
    ev.events = EPOLLIN;
    ev.data.ptr = ptr;

    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, fd,&ev) == -1)
        EXCEPTIONSTREAM(error,t2n_server_error,"error adding server socket to epoll(): " << strerror(errno));

}

/**
 * @brief this removes one file descriptor from the epoll pool
 * assuming the epoll was previously created and the fd was
 * previously added somewhere
 * @param fd the file descriptor we want to monitor
 */
void socket_handler::remove_from_epoll(int fd)
{
    struct epoll_event ev;

    if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, fd,&ev) == -1)
        LOGSTREAM(debug,"error deleting server socket to epoll(): " << strerror(errno) << "  ...ignoring"<<endl);
}

}
